import pytest

from sqlalchemy import exc

from acondbs.db.sa import sa
from acondbs.models import Product, ProductRelation, ProductRelationType

# __________________________________________________________________||
def test_delete_type(app):
    """assert ProductRelationType can only be deleted if relations of the
       type don't exist

    """


    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+

    # attempt to delete the relation type "parent"
    with app.app_context():
        relation_type_parent = ProductRelationType.query.filter_by(name='parent').one_or_none()
        sa.session.delete(relation_type_parent)
        with pytest.raises(exc.IntegrityError):
            sa.session.commit()

    # assert the relation type and the relations of this type still exist
    with app.app_context():
        relation_type_parent = ProductRelationType.query.filter_by(name='parent').one_or_none()
        assert relation_type_parent is not None
        assert 2 == len(relation_type_parent.relations)

    # delete the relations "parent"
    with app.app_context():
        relation_type_parent = ProductRelationType.query.filter_by(name='parent').one_or_none()
        for r in relation_type_parent.relations:
            sa.session.delete(r)
        sa.session.commit()

    # assert the relations "parent" and their reverse relations "child" are deleted.
    # assert the relation types still exist
    with app.app_context():
        relation_type_parent = ProductRelationType.query.filter_by(name='parent').one_or_none()
        assert relation_type_parent is not None
        assert 0 == len(relation_type_parent.relations)
        relation_type_child = ProductRelationType.query.filter_by(name='child').one_or_none()
        assert relation_type_child is not None
        assert 0 == len(relation_type_child.relations)

    # now, delete the relations type "parent"
    with app.app_context():
        relation_type_parent = ProductRelationType.query.filter_by(name='parent').one_or_none()
        sa.session.delete(relation_type_parent)
        sa.session.commit()

    # assert the relation types, both "parent" and its reverse "child", are deleted
    with app.app_context():
        relation_type_parent = ProductRelationType.query.filter_by(name='parent').one_or_none()
        assert relation_type_parent is None
        relation_type_child = ProductRelationType.query.filter_by(name='child').one_or_none()
        assert relation_type_child is None

    # assert Products still exist
    with app.app_context():
        parent1 = Product.query.filter_by(name='parent1').one_or_none()
        child1 = Product.query.filter_by(name='child1').one_or_none()
        child2 = Product.query.filter_by(name='child2').one_or_none()

        assert parent1 is not None
        assert child1 is not None
        assert child2 is not None

# __________________________________________________________________||



import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product

# __________________________________________________________________||
@pytest.fixture
def app(app):
    y = app

    type_map = ProductType(name='map')
    map1 = Product(name="map1", contact='xyz', type_=type_map)

    with y.app_context():
        sa.session.add(type_map)
        sa.session.commit()

    yield y

# __________________________________________________________________||
def test_simple(app):
    '''A simple test of updating an object
    '''

    with app.app_context():
        product1 = Product.query.filter_by(name='map1').first()
        assert 'xyz' == product1.contact
        product1.contact = 'abc'
        sa.session.commit()

    with app.app_context():
        product1 = Product.query.filter_by(name='map1').first()
        assert 'abc' == product1.contact

# __________________________________________________________________||



import graphene

from .query import Query
from .mutation import Mutation

##__________________________________________________________________||
def create_schema(enable_mutation=True):
    if enable_mutation:
        return graphene.Schema(query=Query, mutation=Mutation)
    return graphene.Schema(query=Query)

##__________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

from ..gql import FRAGMENT_PRODUCT_TYPE_CONNECTION

##__________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          { allProductTypes(sort: ORDER_ASC) {
            ...fragmentProductTypeConnection
          }}
         ''' + FRAGMENT_PRODUCT_TYPE_CONNECTION)],
        {},
        id='sort-order'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProductTypes {
              totalCount
            }
          }
        '''),],
        {},
        id='total-count'
    ),
]

@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

# __________________________________________________________________||



from graphene import Int, NonNull
from graphene.relay import Connection


##__________________________________________________________________||
class CountedConnection(Connection):
    '''

    copied from https://github.com/graphql-python/graphene-sqlalchemy/issues/153#issuecomment-501949939

    '''

    class Meta:
        # Being abstract is important because we can't reference the
        # node type here without creating a circular reference. Also, it
        # makes reuse easy.
        #
        # The node type will be populated later with
        # `CountedConnection.create_class()` in `Foo`'s
        # `__init_subclass_with_meta__()`.
        abstract = True

    total_count = NonNull(Int)

    def resolve_total_count(self, info, **kwargs):
        return self.length


##__________________________________________________________________||



from acondbs.models import Product, ProductRelation, ProductRelationType

# __________________________________________________________________||
def test_example_how_to_query(app):

    #                              +--------+
    #               --(child)-->   |        |
    #                    |         | child1 |
    #  +---------+  <-(parent)--   |        |
    #  |         |                 +--------+
    #  | parent1 |
    #  |         |                 +--------+
    #  +---------+  --(child)-->   |        |
    #                    |         | child2 |
    #               <-(parent)--   |        |
    #                              +--------+


    with app.app_context():

        parent1 = Product.query.filter_by(name='parent1').first()
        child1 = Product.query.filter_by(name='child1').first()
        child2 = Product.query.filter_by(name='child2').first()

        # Product relations with the type name "child"
        relations = ProductRelation.query.join(
            ProductRelationType
        ).filter(
            ProductRelationType.name=='child'
        ).all()
        assert ['child', 'child'] == [r.type_.name for r in relations]

        # Products with "parent"
        products = Product.query.join(
            ProductRelation,
            (Product.product_id == ProductRelation.self_product_id)
        ).join(
            ProductRelationType
        ).filter(
            ProductRelationType.name=='parent'
        ).all()
        assert [child1, child2] == products

        # Products with "child"
        products = Product.query.join(
            ProductRelation,
            (Product.product_id == ProductRelation.self_product_id)
        ).join(
            ProductRelationType
        ).filter(
            ProductRelationType.name=='child'
        ).all()
        assert [parent1] == products

# __________________________________________________________________||



import graphene
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import WebConfig as WebConfigModel

from .connection import CountedConnection
from .filter_ import PFilterableConnectionField

##__________________________________________________________________||
class WebConfig(SQLAlchemyObjectType):
    '''Web configuration'''
    class Meta:
        model = WebConfigModel
        interfaces = (graphene.relay.Node, )
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory

##__________________________________________________________________||



"""declare ORM models

In this packages, ORM (Object-relational mapping) models are declared.
One model is mapped to one table in the DB. Models are declared as
Python classes inheriting the Model class in Flask-SQLAlchemy.

"Declaring Models" in Flask-SQLAlchemy doc:
https://flask-sqlalchemy.palletsprojects.com/en/2.x/models/

"Declare a Mapping" in SQLAlchemy doc:
https://docs.sqlalchemy.org/en/13/orm/tutorial.html#declare-a-mapping

"""

##__________________________________________________________________||
from .product_type import ProductType
from .product import Product
from .product_file_path import ProductFilePath
from .product_relation_type import ProductRelationType
from .product_relation import ProductRelation

from .web_config import WebConfig

from .admin_app_token import AdminAppToken

##__________________________________________________________________||



from ..db.sa import sa

##__________________________________________________________________||
class Product(sa.Model):
    __tablename__ = 'products'
    product_id = sa.Column(sa.Integer(), primary_key=True)
    type_id = sa.Column(
        sa.ForeignKey('product_types.type_id'),
        nullable=False
    )
    type_ = sa.relationship("ProductType", backref=sa.backref("products"))
    name = sa.Column(sa.Text(), nullable=False)
    contact = sa.Column(sa.Text())
    date_produced = sa.Column(sa.Date())
    produced_by = sa.Column(sa.Text())
    date_posted = sa.Column(sa.Date())
    posted_by = sa.Column(sa.Text())
    date_updated = sa.Column(sa.Date())
    updated_by = sa.Column(sa.Text())
    note = sa.Column(sa.Text())
    __table_args__ = (sa.UniqueConstraint('type_id', 'name', name='_type_id_name'), )

##__________________________________________________________________||



import graphene
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import Product as ProductModel
from ..models import ProductType as ProductTypeModel

from ..db.sa import sa
from ..db.backup import request_backup_db

from .connection import CountedConnection
from .filter_ import PFilterableConnectionField

##__________________________________________________________________||
class ProductType(SQLAlchemyObjectType):
    '''A product type'''
    class Meta:
        model = ProductTypeModel
        interfaces = (graphene.relay.Node, )
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory

##__________________________________________________________________||
class CommonInputFields:
    order = graphene.Int(
        description=('The order in which the type is displayed, for example, '
                     'in navigation bars.'))
    indef_article = graphene.String(
        description=('The indefinite article placed before the singular noun "'
                     'i.e., "a" or "an". '))
    singular = graphene.String(
        description=('The singular noun, the product type name in singular.'))
    plural = graphene.String(
        description=('The plural noun, the product type name in plural.'))
    icon = graphene.String(
        description=('A name of the icon from https://materialdesignicons.com/'))

class CreateProductTypeInput(graphene.InputObjectType, CommonInputFields):
    '''Input to createProductType()'''
    name = graphene.String(required=True, description='The name of the product type')

class UpdateProductTypeInput(graphene.InputObjectType,CommonInputFields):
    '''Input to updateProductType()'''

##__________________________________________________________________||
class CreateProductType(graphene.Mutation):
    '''Create a product type'''
    class Arguments:
        input = CreateProductTypeInput(required=True)

    ok = graphene.Boolean()
    product_type = graphene.Field(lambda: ProductType)

    def mutate(root, info, input):
        model = ProductTypeModel(**input)
        sa.session.add(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return CreateProductType(product_type=model, ok=ok)

class UpdateProductType(graphene.Mutation):
    '''Update a product type'''
    class Arguments:
        type_id = graphene.Int(required=True)
        input = UpdateProductTypeInput(required=True)

    ok = graphene.Boolean()
    product_type = graphene.Field(lambda: ProductType)

    def mutate(root, info, type_id, input):
        model = ProductTypeModel.query.filter_by(type_id=type_id).one()
        for k, v in input.items():
            setattr(model, k, v)
        sa.session.commit()
        ok = True
        request_backup_db()
        return UpdateProductType(product_type=model, ok=ok)

class DeleteProductType(graphene.Mutation):
    '''Delete a product type'''
    class Arguments:
        type_id = graphene.Int(description='The typeId of the product type')

    ok = graphene.Boolean()

    def mutate(root, info, type_id):
        model = ProductTypeModel.query.filter_by(type_id=type_id).one()
        sa.session.delete(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteProductType(ok=ok)

##__________________________________________________________________||



from ..db.sa import sa

from sqlalchemy.event import listens_for

##__________________________________________________________________||
class ProductRelation(sa.Model):
    __tablename__ = 'product_relations'
    relation_id = sa.Column(sa.Integer(), primary_key=True)
    type_id = sa.Column(
        sa.Integer(),
        sa.ForeignKey('product_relation_types.type_id'),
        nullable=False)
    type_ = sa.relationship(
        "ProductRelationType",
        backref=sa.backref("relations")
    )
    self_product_id = sa.Column(
        sa.Integer(),
        sa.ForeignKey('products.product_id'),
        nullable=False)
    self_ = sa.relationship(
        "Product",
        foreign_keys=[self_product_id],
        backref=sa.backref("relations", cascade="all, delete-orphan"))
    other_product_id = sa.Column(
        sa.Integer(),
        sa.ForeignKey('products.product_id'),
        nullable=False)
    other = sa.relationship("Product", foreign_keys=[other_product_id])
    reverse_relation_id = sa.Column(sa.ForeignKey('product_relations.relation_id'))
    reverse = sa.relationship(
        lambda: ProductRelation,
        uselist=False,
        foreign_keys=[reverse_relation_id],
        remote_side="ProductRelation.relation_id",
        cascade="all",
        post_update=True)
    __table_args__ = (
        sa.UniqueConstraint(
            'type_id', 'self_product_id', 'other_product_id',
            name='_type_id_self_product_id_other_product_id'), )

##__________________________________________________________________||
@listens_for(ProductRelation.type_, 'set')
def set_reverse_type(target, value, oldvalue, initiator):
    relation = target

    try:
        if relation.__avoid_recursive:
            return
    except:
        pass

    if not value.reverse:
        return

    if not relation.reverse:
        reverse = ProductRelation()
        reverse.reverse = relation
        relation.reverse = reverse
        reverse.__avoid_recursive = True
        if relation.self_:
            reverse.other = relation.self_
        if relation.other:
            reverse.self_ = relation.other
        del reverse.__avoid_recursive
    if not relation.reverse.type_:
        relation.reverse.__avoid_recursive = True
        relation.reverse.type_ = value.reverse
        del relation.reverse.__avoid_recursive

@listens_for(ProductRelation.self_, 'set')
def set_reverse_other(target, value, oldvalue, initiator):
    relation = target

    if not relation.reverse:
        return

    try:
        if relation.__avoid_recursive:
            return
    except:
        pass

    if not relation.reverse.other:
        relation.reverse.__avoid_recursive = True
        relation.reverse.other = value
        del relation.reverse.__avoid_recursive

@listens_for(ProductRelation.other, 'set')
def set_reverse_self(target, value, oldvalue, initiator):
    relation = target

    if not relation.reverse:
        return

    try:
        if relation.__avoid_recursive:
            return
    except:
        pass

    if not relation.reverse.self_:
        relation.reverse.__avoid_recursive = True
        relation.reverse.self_ = value
        del relation.reverse.__avoid_recursive

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[type_id] 1'] = {
    'data': {
        'productRelation': {
            'other': {
                'name': 'beam1',
                'productId': '4'
            },
            'relationId': '1',
            'reverse': {
                'relationId': '2'
            },
            'self_': {
                'name': 'map1',
                'productId': '1'
            },
            'type_': {
                'name': 'child',
                'typeId': '2'
            }
        }
    }
}

snapshots['test_schema[type_id-nonexistent] 1'] = {
    'data': {
        'productRelation': None
    }
}



import pytest

from ..funcs import assert_mutation

from ..gql import (
    FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION,
    CREATE_PRODUCT_RELATION_TYPES
    )

##__________________________________________________________________||
QEURY = '''
{
  allProductRelationTypes {
   ...fragmentProductRelationTypeConnection
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION_TYPES],
            {
                'variables': {
                    'type': {
                        'name': "doctor",
                        'indefArticle': "a",
                        'singular': "doctor",
                        'plural': "doctors"
                    },
                    'reverse': {
                        'name': "patient",
                        'indefArticle': "a",
                        'singular': "patient",
                        'plural': "patients"
                    },
                }}
        ],
        [[QEURY], {}],
        id='reverse'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION_TYPES],
            {
                'variables': {
                    'type': {
                        'name': "spouse",
                        'indefArticle': "a",
                        'singular': "spouse",
                        'plural': "spouses"
                    },
                    'selfReverse': True
                    }
            }
        ],
        [[QEURY], {}],
        id='self_reverse'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)


##__________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION_TYPES],
            {
                'variables': {
                    'type': {
                        'name': "parent",
                        'indefArticle': "a",
                        'singular': "parent",
                        'plural': "parents"
                    },
                    'reverse': {
                        'name': "child",
                        'indefArticle': "a",
                        'singular': "child",
                        'plural': "children"
                    }
                }
            },
        ],
        [[QEURY], {}],
        id='error-already-exist'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION_TYPES],
            {
                'variables': {
                    'type': {
                        'name': "doctor",
                        'indefArticle': "a",
                        'singular': "doctor",
                        'plural': "doctors"
                    },
                    'reverse': {
                        'name': "patient",
                        'indefArticle': "a",
                        'singular': "patient",
                        'plural': "patients"
                    },
                    'selfReverse': True
                }
            }
        ],
        [[QEURY], {}],
        id='error-reverse-and-self_reverse'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

##__________________________________________________________________||



import graphene

from .product import (
    CreateProduct, UpdateProduct, DeleteProduct)
from .product_file_path import (
    CreateProductFilePath, UpdateProductFilePath, DeleteProductFilePath)
from .product_type import (
    CreateProductType, UpdateProductType, DeleteProductType)
from .product_relation_type import (
    CreateProductRelationTypes, UpdateProductRelationType, DeleteProductRelationTypes)
from .product_relation import (
    CreateProductRelation, DeleteProductRelation)

from .auth import GitHubAuth, StoreAdminAppToken

from .query import Query

##__________________________________________________________________||
class Mutation(graphene.ObjectType):

    create_product = CreateProduct.Field()
    update_product = UpdateProduct.Field()
    delete_product = DeleteProduct.Field()

    create_product_file_path = CreateProductFilePath.Field()
    update_product_file_path = UpdateProductFilePath.Field()
    delete_product_file_path = DeleteProductFilePath.Field()

    create_product_type = CreateProductType.Field()
    update_product_type = UpdateProductType.Field()
    delete_product_type = DeleteProductType.Field()

    create_product_relation_types = CreateProductRelationTypes.Field()
    update_product_relation_type = UpdateProductRelationType.Field()
    delete_product_relation_types = DeleteProductRelationTypes.Field()

    create_product_relation = CreateProductRelation.Field()
    delete_product_relation = DeleteProductRelation.Field()

    github_auth = GitHubAuth.Field()
    store_admin_app_token = StoreAdminAppToken.Field()

##__________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

from ..gql import (
    FRAGMENT_PRODUCT_CONNECTION_SHALLOW,
    FRAGMENT_PRODUCT_CONNECTION_DEEP
    )

##__________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts {
              ...fragmentProductConnectionDeep
            }
          }
        ''') + FRAGMENT_PRODUCT_CONNECTION_DEEP,],
        {},
        id='deep'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts(first: 2) {
              ...fragmentProductConnectionShallow
            }
          }
        ''') + FRAGMENT_PRODUCT_CONNECTION_SHALLOW,],
        {},
        id='first-two'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts(first: 2, sort: DATE_PRODUCED_DESC) {
              ...fragmentProductConnectionShallow
            }
          }
         ''') + FRAGMENT_PRODUCT_CONNECTION_SHALLOW,],
        {},
        id='first-two-sort'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts(filters: {typeId: 1}, first: 2) {
              ...fragmentProductConnectionShallow
            }
          }
         ''') + FRAGMENT_PRODUCT_CONNECTION_SHALLOW,],
        {},
        id='filters-type_id-first-two'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts(filters: {typeName: "map"}, first: 2) {
              ...fragmentProductConnectionShallow
            }
          }
         ''') + FRAGMENT_PRODUCT_CONNECTION_SHALLOW,],
        {},
        id='filters-type_name-first-two'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts(filters: {typeId: 1}, sort: DATE_PRODUCED_DESC) {
              ...fragmentProductConnectionShallow
            }
          }
         ''') + FRAGMENT_PRODUCT_CONNECTION_SHALLOW,],
        {},
        id='filters-type_id-sort'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts(filters: {typeName: "map"}, sort: DATE_PRODUCED_DESC) {
              ...fragmentProductConnectionShallow
            }
          }
         ''') + FRAGMENT_PRODUCT_CONNECTION_SHALLOW,],
        {},
        id='filters-type_name-sort'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            allProducts(filters: {typeName: "map"}, first: 2) {
              totalCount
            }
          }
         '''),],
        {},
        id='total-count'
    ),
]

##__________________________________________________________________||
@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_success[create] 1'] = {
    'data': {
        'createProduct': {
            'ok': True,
            'product': {
                'contact': 'contact-person',
                'datePosted': '2020-05-04',
                'dateProduced': '2020-02-20',
                'dateUpdated': None,
                'name': 'beam111',
                'note': '- Item 1',
                'paths': {
                    'edges': [
                        {
                            'node': {
                                'note': None,
                                'path': '/path/to/new/product1',
                                'pathId': '9'
                            }
                        },
                        {
                            'node': {
                                'note': None,
                                'path': '/another/location/of/product1',
                                'pathId': '10'
                            }
                        }
                    ]
                },
                'postedBy': 'poster',
                'producedBy': 'producer',
                'productId': '6',
                'relations': {
                    'edges': [
                        {
                            'node': {
                                'other': {
                                    'name': 'map1',
                                    'productId': '1',
                                    'typeId': 1,
                                    'type_': {
                                        'name': 'map',
                                        'typeId': '1'
                                    }
                                },
                                'otherProductId': 1,
                                'relationId': '8',
                                'reverse': {
                                    'relationId': '7',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'child',
                                        'typeId': '2'
                                    }
                                },
                                'reverseRelationId': 7,
                                'typeId': 1,
                                'type_': {
                                    'name': 'parent',
                                    'typeId': '1'
                                }
                            }
                        },
                        {
                            'node': {
                                'other': {
                                    'name': 'beam2',
                                    'productId': '5',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'beam',
                                        'typeId': '2'
                                    }
                                },
                                'otherProductId': 5,
                                'relationId': '10',
                                'reverse': {
                                    'relationId': '9',
                                    'typeId': 2,
                                    'type_': {
                                        'name': 'child',
                                        'typeId': '2'
                                    }
                                },
                                'reverseRelationId': 9,
                                'typeId': 1,
                                'type_': {
                                    'name': 'parent',
                                    'typeId': '1'
                                }
                            }
                        }
                    ]
                },
                'typeId': 2,
                'type_': {
                    'name': 'beam',
                    'typeId': '2'
                },
                'updatedBy': None
            }
        }
    }
}

snapshots['test_schema_success[create] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': '/path/to/new/product1'
                    }
                },
                {
                    'node': {
                        'path': '/another/location/of/product1'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam111'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam111'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam111'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam111'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                },
                {
                    'node': {
                        'name': 'beam111'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_error[error-no-name] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[minimum] 1'] = {
    'data': {
        'createProduct': {
            'ok': True,
            'product': {
                'contact': None,
                'datePosted': '2020-05-04',
                'dateProduced': None,
                'dateUpdated': None,
                'name': 'product1',
                'note': None,
                'paths': {
                    'edges': [
                    ]
                },
                'postedBy': None,
                'producedBy': None,
                'productId': '6',
                'relations': {
                    'edges': [
                    ]
                },
                'typeId': 1,
                'type_': {
                    'name': 'map',
                    'typeId': '1'
                },
                'updatedBy': None
            }
        }
    }
}

snapshots['test_schema_success[minimum] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                },
                {
                    'node': {
                        'name': 'product1'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[the-same-name-different-type] 1'] = {
    'data': {
        'createProduct': {
            'ok': True,
            'product': {
                'contact': 'contact-person',
                'datePosted': '2020-05-04',
                'dateProduced': None,
                'dateUpdated': None,
                'name': 'map1',
                'note': None,
                'paths': {
                    'edges': [
                        {
                            'node': {
                                'note': None,
                                'path': '/path/to/new/product1',
                                'pathId': '9'
                            }
                        },
                        {
                            'node': {
                                'note': None,
                                'path': '/another/location/of/product1',
                                'pathId': '10'
                            }
                        }
                    ]
                },
                'postedBy': None,
                'producedBy': 'pwg-pmn',
                'productId': '6',
                'relations': {
                    'edges': [
                    ]
                },
                'typeId': 2,
                'type_': {
                    'name': 'beam',
                    'typeId': '2'
                },
                'updatedBy': None
            }
        }
    }
}

snapshots['test_schema_success[the-same-name-different-type] 2'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                },
                {
                    'node': {
                        'path': '/path/to/new/product1'
                    }
                },
                {
                    'node': {
                        'path': '/another/location/of/product1'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                },
                {
                    'node': {
                        'name': 'map1'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_error[error-the-same-type-and-name] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'path': 'site1:/path/to/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/beam1'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map2'
                    }
                },
                {
                    'node': {
                        'path': 'site1:/path/to/map3'
                    }
                },
                {
                    'node': {
                        'path': 'site2:/another/way/map3'
                    }
                }
            ]
        },
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'beam1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1'
                        },
                        'self_': {
                            'name': 'beam2'
                        },
                        'type_': {
                            'name': 'parent'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2'
                        },
                        'self_': {
                            'name': 'map1'
                        },
                        'type_': {
                            'name': 'child'
                        }
                    }
                }
            ]
        },
        'allProducts': {
            'edges': [
                {
                    'node': {
                        'name': 'map1'
                    }
                },
                {
                    'node': {
                        'name': 'map2'
                    }
                },
                {
                    'node': {
                        'name': 'map3'
                    }
                },
                {
                    'node': {
                        'name': 'beam1'
                    }
                },
                {
                    'node': {
                        'name': 'beam2'
                    }
                }
            ]
        }
    }
}




##__________________________________________________________________||
FRAGMENT_PRODUCT_RELATION = '''
fragment fragmentProductRelation on ProductRelation {
  relationId
  type_ {
    typeId
    name
  }
  self_ {
    productId
    name
  }
  other {
    productId
    name
  }
  reverse {
    relationId
  }
}
'''

FRAGMENT_PRODUCT_RELATION_CONNECTION = '''
fragment fragmentProductRelationConnection on ProductRelationConnection {
  edges {
    node {
      ...fragmentProductRelation
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION

##__________________________________________________________________||
CREATE_PRODUCT_RELATION = '''
mutation CreateProductRelation($input: CreateProductRelationInput!) {
  createProductRelation(input: $input) {
    ok
    productRelation {
      ...fragmentProductRelation
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION

DELETE_PRODUCT_RELATION = '''
mutation DeleteProductRelation($relationId: Int!) {
  deleteProductRelation(relationId: $relationId) {
    ok
  }
}
'''

##__________________________________________________________________||



"""empty message

Revision ID: 0cce33011766
Revises: 3f4e8bfd8a7f
Create Date: 2020-05-25 19:22:24.114156

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '0cce33011766'
down_revision = '3f4e8bfd8a7f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('product_relations', schema=None) as batch_op:
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('products', schema=None) as batch_op:
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    with op.batch_alter_table('product_relations', schema=None) as batch_op:
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=True)

    # ### end Alembic commands ###



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import GenericRepr, Snapshot


snapshots = Snapshot()

snapshots['test_define_tables_start_with_empty_db 1'] = {
    'admin_app_token': GenericRepr("Table('admin_app_token', MetaData(bind=None), Column('token_id', INTEGER(), table=<admin_app_token>, primary_key=True, nullable=False), Column('token', BLOB(), table=<admin_app_token>), schema=None)"),
    'product_file_paths': GenericRepr("Table('product_file_paths', MetaData(bind=None), Column('path_id', INTEGER(), table=<product_file_paths>, primary_key=True, nullable=False), Column('path', TEXT(), table=<product_file_paths>), Column('note', TEXT(), table=<product_file_paths>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_file_paths>), schema=None)"),
    'product_relation_types': GenericRepr("Table('product_relation_types', MetaData(bind=None), Column('type_id', INTEGER(), table=<product_relation_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_relation_types>, nullable=False), Column('reverse_type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relation_types>), Column('indef_article', TEXT(), table=<product_relation_types>), Column('singular', TEXT(), table=<product_relation_types>), Column('plural', TEXT(), table=<product_relation_types>), schema=None)"),
    'product_relations': GenericRepr("Table('product_relations', MetaData(bind=None), Column('relation_id', INTEGER(), table=<product_relations>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relations>, nullable=False), Column('self_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('other_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('reverse_relation_id', INTEGER(), ForeignKey('product_relations.relation_id'), table=<product_relations>), schema=None)"),
    'product_types': GenericRepr("Table('product_types', MetaData(bind=None), Column('type_id', INTEGER(), table=<product_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_types>, nullable=False), Column('order', INTEGER(), table=<product_types>), Column('indef_article', TEXT(), table=<product_types>), Column('singular', TEXT(), table=<product_types>), Column('plural', TEXT(), table=<product_types>), Column('icon', TEXT(), table=<product_types>), schema=None)"),
    'products': GenericRepr("Table('products', MetaData(bind=None), Column('product_id', INTEGER(), table=<products>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_types.type_id'), table=<products>, nullable=False), Column('name', TEXT(), table=<products>, nullable=False), Column('contact', TEXT(), table=<products>), Column('date_produced', DATE(), table=<products>), Column('produced_by', TEXT(), table=<products>), Column('date_posted', DATE(), table=<products>), Column('posted_by', TEXT(), table=<products>), Column('date_updated', DATE(), table=<products>), Column('updated_by', TEXT(), table=<products>), Column('note', TEXT(), table=<products>), schema=None)"),
    'web_config': GenericRepr("Table('web_config', MetaData(bind=None), Column('config_id', INTEGER(), table=<web_config>, primary_key=True, nullable=False), Column('head_title', TEXT(), table=<web_config>), Column('toolbar_title', TEXT(), table=<web_config>), Column('devtool_loadingstate', BOOLEAN(), table=<web_config>), Column('product_creation_dialog', BOOLEAN(), table=<web_config>), Column('product_update_dialog', BOOLEAN(), table=<web_config>), Column('product_deletion_dialog', BOOLEAN(), table=<web_config>), schema=None)")
}

snapshots['test_define_tables_start_with_nonempty_db 1'] = {
    'admin_app_token': GenericRepr("Table('admin_app_token', MetaData(bind=None), Column('token_id', INTEGER(), table=<admin_app_token>, primary_key=True, nullable=False), Column('token', BLOB(), table=<admin_app_token>), schema=None)"),
    'product_file_paths': GenericRepr("Table('product_file_paths', MetaData(bind=None), Column('path_id', INTEGER(), table=<product_file_paths>, primary_key=True, nullable=False), Column('path', TEXT(), table=<product_file_paths>), Column('note', TEXT(), table=<product_file_paths>), Column('product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_file_paths>), schema=None)"),
    'product_relation_types': GenericRepr("Table('product_relation_types', MetaData(bind=None), Column('type_id', INTEGER(), table=<product_relation_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_relation_types>, nullable=False), Column('reverse_type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relation_types>), Column('indef_article', TEXT(), table=<product_relation_types>), Column('singular', TEXT(), table=<product_relation_types>), Column('plural', TEXT(), table=<product_relation_types>), schema=None)"),
    'product_relations': GenericRepr("Table('product_relations', MetaData(bind=None), Column('relation_id', INTEGER(), table=<product_relations>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_relation_types.type_id'), table=<product_relations>, nullable=False), Column('self_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('other_product_id', INTEGER(), ForeignKey('products.product_id'), table=<product_relations>, nullable=False), Column('reverse_relation_id', INTEGER(), ForeignKey('product_relations.relation_id'), table=<product_relations>), schema=None)"),
    'product_types': GenericRepr("Table('product_types', MetaData(bind=None), Column('type_id', INTEGER(), table=<product_types>, primary_key=True, nullable=False), Column('name', TEXT(), table=<product_types>, nullable=False), Column('order', INTEGER(), table=<product_types>), Column('indef_article', TEXT(), table=<product_types>), Column('singular', TEXT(), table=<product_types>), Column('plural', TEXT(), table=<product_types>), Column('icon', TEXT(), table=<product_types>), schema=None)"),
    'products': GenericRepr("Table('products', MetaData(bind=None), Column('product_id', INTEGER(), table=<products>, primary_key=True, nullable=False), Column('type_id', INTEGER(), ForeignKey('product_types.type_id'), table=<products>, nullable=False), Column('name', TEXT(), table=<products>, nullable=False), Column('contact', TEXT(), table=<products>), Column('date_produced', DATE(), table=<products>), Column('produced_by', TEXT(), table=<products>), Column('date_posted', DATE(), table=<products>), Column('posted_by', TEXT(), table=<products>), Column('date_updated', DATE(), table=<products>), Column('updated_by', TEXT(), table=<products>), Column('note', TEXT(), table=<products>), schema=None)"),
    'web_config': GenericRepr("Table('web_config', MetaData(bind=None), Column('config_id', INTEGER(), table=<web_config>, primary_key=True, nullable=False), Column('head_title', TEXT(), table=<web_config>), Column('toolbar_title', TEXT(), table=<web_config>), Column('devtool_loadingstate', BOOLEAN(), table=<web_config>), Column('product_creation_dialog', BOOLEAN(), table=<web_config>), Column('product_update_dialog', BOOLEAN(), table=<web_config>), Column('product_deletion_dialog', BOOLEAN(), table=<web_config>), schema=None)")
}



from ..db.sa import sa

##__________________________________________________________________||
class ProductType(sa.Model):
    __tablename__ = 'product_types'
    type_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    order = sa.Column(sa.Integer())
    indef_article = sa.Column(sa.Text())
    singular = sa.Column(sa.Text())
    plural = sa.Column(sa.Text())
    icon = sa.Column(sa.Text())

##__________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

from ..gql import (
    FRAGMENT_PRODUCT_SHALLOW,
    FRAGMENT_PRODUCT_DEEP
)

# __________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          query Product($productId: Int) {
            product(productId: $productId) {
              ...fragmentProductDeep
            }
          }
          ''') + FRAGMENT_PRODUCT_DEEP, ],
        {'variables': {'productId': 1}},
        id='deep'
    ),
    pytest.param(
        [textwrap.dedent('''
          query Product($productId: Int) {
            product(productId: $productId) {
              ...fragmentProductDeep
            }
          }
          ''') + FRAGMENT_PRODUCT_DEEP, ],
        {'variables': {'productId': 9899}},
        id='product_id-nonexistent'
    ),
    pytest.param(
        [textwrap.dedent('''
          query ProductByTypeIdAndName($typeId: Int!, $name: String!) {
            product(typeId: $typeId, name: $name) {
              ...fragmentProductShallow
            }
          }
        ''') + FRAGMENT_PRODUCT_SHALLOW, ],
        {'variables': {'typeId': 1, 'name': "map1"}},
        id='type_id-name'
    ),
    pytest.param(
        [textwrap.dedent('''
        { product(productId: 1, name: "map1") {
            ...fragmentProductShallow
          }
        }
         ''') + FRAGMENT_PRODUCT_SHALLOW, ],
        {},
        id='product_id-name'
    ),
    pytest.param(
        [textwrap.dedent('''
        { product(productId: 1, name: "map2") {
            ...fragmentProductShallow
          }
        }
         ''') + FRAGMENT_PRODUCT_SHALLOW, ],
        {},
        id='product_id-name-nonexistent'
    ),
]

# __________________________________________________________________||


@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

# __________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[sort-order] 1'] = {
    'data': {
        'allProductTypes': {
            'edges': [
                {
                    'node': {
                        'icon': 'mdi-spotlight-beam',
                        'indefArticle': 'a',
                        'name': 'beam',
                        'order': 1,
                        'plural': 'beams',
                        'products': {
                            'edges': [
                            ]
                        },
                        'singular': 'beam',
                        'typeId': '2'
                    }
                },
                {
                    'node': {
                        'icon': 'mdi-map',
                        'indefArticle': 'a',
                        'name': 'map',
                        'order': 2,
                        'plural': 'maps',
                        'products': {
                            'edges': [
                                {
                                    'node': {
                                        'name': 'map1'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map2'
                                    }
                                },
                                {
                                    'node': {
                                        'name': 'map3'
                                    }
                                }
                            ]
                        },
                        'singular': 'map',
                        'typeId': '1'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[total-count] 1'] = {
    'data': {
        'allProductTypes': {
            'totalCount': 2
        }
    }
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[all] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'note': '',
                        'path': 'nersc:/go/to/my/maps',
                        'pathId': '1'
                    }
                },
                {
                    'node': {
                        'note': 'lat only',
                        'path': 'nersc:/go/to/my/maps_v2',
                        'pathId': '2'
                    }
                },
                {
                    'node': {
                        'note': 'lat only',
                        'path': 'abcde:/path/to/the/maps_v2',
                        'pathId': '3'
                    }
                },
                {
                    'node': {
                        'note': 'lat only',
                        'path': 'nersc:/go/to/my/maps_v3',
                        'pathId': '4'
                    }
                },
                {
                    'node': {
                        'note': '',
                        'path': 'BEAM_DEPOT/Beams/20190304',
                        'pathId': '5'
                    }
                },
                {
                    'node': {
                        'note': '',
                        'path': 'BEAM_DEPOT/Beams/20190607',
                        'pathId': '6'
                    }
                },
                {
                    'node': {
                        'note': '',
                        'path': 'BEAM_DEPOT/Beams/20200123',
                        'pathId': '7'
                    }
                },
                {
                    'node': {
                        'note': '',
                        'path': 'BEAM_DEPOT/Beams/20200207',
                        'pathId': '8'
                    }
                },
                {
                    'node': {
                        'note': '',
                        'path': 'nersc:/go/to/my/simulations',
                        'pathId': '9'
                    }
                },
                {
                    'node': {
                        'note': '',
                        'path': 'abcde:/path/to/the/simulations',
                        'pathId': '10'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[total-count] 1'] = {
    'data': {
        'allProductFilePaths': {
            'totalCount': 10
        }
    }
}



import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product, ProductFilePath

##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri ="sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y


@pytest.fixture
def app(app_empty):

    y = app_empty

    # product --- path1
    #          |
    #          +- path2
    type1 = ProductType(name='type1')
    product = Product(name="product", type_=type1)
    path1 = ProductFilePath(path="/path1")
    path2 = ProductFilePath(path="/path2")
    product.paths = [path1, path2]

    # commit
    with y.app_context():
        sa.session.add(product)
        sa.session.commit()
    yield y

# __________________________________________________________________||
def test_relations(app):

    with app.app_context():
        product = Product.query.filter_by(name='product').one_or_none()
        assert 'product' == product.name
        assert 2 == len(product.paths)
        path1, path2 = product.paths
        assert '/path1' == path1.path
        assert '/path2' == path2.path
        assert product is path1.product
        assert product is path2.product

def test_delete_product(app):

    with app.app_context():
        product = Product.query.filter_by(name='product').one_or_none()
        sa.session.delete(product)
        sa.session.commit()

    with app.app_context():
        product = Product.query.filter_by(name='product').one_or_none()
        assert product is None
        assert [] == ProductFilePath.query.all()

def test_delete_path(app):

    with app.app_context():
        path1 = ProductFilePath.query.filter_by(path='/path1').one_or_none()
        sa.session.delete(path1)
        sa.session.commit()

    with app.app_context():
        product = Product.query.filter_by(name='product').one_or_none()
        assert product is not None
        path2 = ProductFilePath.query.filter_by(path='/path2').one_or_none()
        assert [path2] == ProductFilePath.query.all()

        sa.session.delete(path2)
        sa.session.commit()

    with app.app_context():
        product = Product.query.filter_by(name='product').one_or_none()
        assert product is not None
        assert [] == ProductFilePath.query.all()

# __________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[type_id] 1'] = {
    'data': {
        'productRelationType': {
            'indefArticle': 'a',
            'name': 'parent',
            'plural': 'parents',
            'relations': {
                'edges': [
                    {
                        'node': {
                            'other': {
                                'name': 'map1',
                                'productId': '1'
                            },
                            'self_': {
                                'name': 'beam2',
                                'productId': '5'
                            }
                        }
                    },
                    {
                        'node': {
                            'other': {
                                'name': 'beam1',
                                'productId': '4'
                            },
                            'self_': {
                                'name': 'beam2',
                                'productId': '5'
                            }
                        }
                    },
                    {
                        'node': {
                            'other': {
                                'name': 'map1',
                                'productId': '1'
                            },
                            'self_': {
                                'name': 'beam1',
                                'productId': '4'
                            }
                        }
                    }
                ]
            },
            'reverse': {
                'name': 'child',
                'typeId': '2'
            },
            'singular': 'parent',
            'typeId': '1'
        }
    }
}

snapshots['test_schema[name] 1'] = {
    'data': {
        'productRelationType': {
            'indefArticle': 'a',
            'name': 'parent',
            'plural': 'parents',
            'relations': {
                'edges': [
                    {
                        'node': {
                            'other': {
                                'name': 'map1',
                                'productId': '1'
                            },
                            'self_': {
                                'name': 'beam2',
                                'productId': '5'
                            }
                        }
                    },
                    {
                        'node': {
                            'other': {
                                'name': 'beam1',
                                'productId': '4'
                            },
                            'self_': {
                                'name': 'beam2',
                                'productId': '5'
                            }
                        }
                    },
                    {
                        'node': {
                            'other': {
                                'name': 'map1',
                                'productId': '1'
                            },
                            'self_': {
                                'name': 'beam1',
                                'productId': '4'
                            }
                        }
                    }
                ]
            },
            'reverse': {
                'name': 'child',
                'typeId': '2'
            },
            'singular': 'parent',
            'typeId': '1'
        }
    }
}

snapshots['test_schema[type_id-and-name] 1'] = {
    'data': {
        'productRelationType': {
            'indefArticle': 'a',
            'name': 'parent',
            'plural': 'parents',
            'relations': {
                'edges': [
                    {
                        'node': {
                            'other': {
                                'name': 'map1',
                                'productId': '1'
                            },
                            'self_': {
                                'name': 'beam2',
                                'productId': '5'
                            }
                        }
                    },
                    {
                        'node': {
                            'other': {
                                'name': 'beam1',
                                'productId': '4'
                            },
                            'self_': {
                                'name': 'beam2',
                                'productId': '5'
                            }
                        }
                    },
                    {
                        'node': {
                            'other': {
                                'name': 'map1',
                                'productId': '1'
                            },
                            'self_': {
                                'name': 'beam1',
                                'productId': '4'
                            }
                        }
                    }
                ]
            },
            'reverse': {
                'name': 'child',
                'typeId': '2'
            },
            'singular': 'parent',
            'typeId': '1'
        }
    }
}

snapshots['test_schema[type_id-and-name-nonexistent)] 1'] = {
    'data': {
        'productRelationType': None
    }
}



import graphene
from graphene import relay
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import ProductFilePath as ProductFilePathModel

from ..db.sa import sa
from ..db.backup import request_backup_db

from .connection import CountedConnection
from .filter_ import PFilterableConnectionField

##__________________________________________________________________||
class ProductFilePath(SQLAlchemyObjectType):
    class Meta:
        model = ProductFilePathModel
        interfaces = (relay.Node, )
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory

##__________________________________________________________________||
class CommonInputFields:
    """Common input fields of mutations for creating and updating file paths

    """
    path = graphene.String()
    note = graphene.String()

class CreateProductFilePathInput(graphene.InputObjectType, CommonInputFields):
    product_id = graphene.Int()

class UpdateProductFilePathInput(graphene.InputObjectType, CommonInputFields):
    pass

##__________________________________________________________________||
class CreateProductFilePath(graphene.Mutation):
    class Arguments:
        input = CreateProductFilePathInput(required=True)

    ok = graphene.Boolean()
    productFilePath = graphene.Field(lambda: ProductFilePath)

    def mutate(root, info, input):
        path = ProductFilePathModel(**input)
        sa.session.add(path)
        sa.session.commit()
        ok = True
        request_backup_db()
        return CreateProductFilePath(productFilePath=path, ok=ok)

class UpdateProductFilePath(graphene.Mutation):
    class Arguments:
        path_id = graphene.Int()
        input = UpdateProductFilePathInput(required=True)

    ok = graphene.Boolean()
    productFilePath = graphene.Field(lambda: ProductFilePath)

    def mutate(root, info, path_id, input):
        path = ProductFilePathModel.query.filter_by(path_id=path_id).first()
        for k, v in input.items():
            setattr(path, k, v)
        sa.session.commit()
        ok = True
        request_backup_db()
        return UpdateProductFilePath(productFilePath=path, ok=ok)

class DeleteProductFilePath(graphene.Mutation):
    class Arguments:
        path_id = graphene.Int()

    ok = graphene.Boolean()

    def mutate(root, info, path_id):
        path = ProductFilePathModel.query.filter_by(path_id=path_id).first()
        sa.session.delete(path)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteProductFilePath(ok=ok)

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [
                '''
                  mutation m {
                    updateProductFilePath(pathId: 1, input: {
                      path: "nersc:/go/to/my/new_product_v2",
                      note: "- Note 1 updated",
                    }) { productFilePath { path } }
                  }
                ''',
                ],
            {}
        ],
        [
            [
                '''
                  {
                    product(productId: 1001) {
                      name datePosted producedBy note
                      paths { edges { node { path note product { productId } } } }
                    }
                  }
                ''',
            ],
            {},
        ],
        id='updateProductFilePath'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [
                '''
                  mutation m {
                    updateProductFilePath(pathId: 1, input: {
                      productId: 1012
                    }) { productFilePath { path } }
                  }
                ''',
            ],
            {},
        ],
        [
            [
                '''
                  {
                    product(productId: 1001) {
                      paths { edges { node { path note product { productId } } } }
                    }
                  }
                ''',
            ],
            {},
        ],
        id='updateProductFilePath-immutableField'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

# __________________________________________________________________||



from graphene.test import Client
from graphene import Context
from werkzeug.datastructures import Headers

import pytest
import unittest.mock as mock

from acondbs.schema.schema import create_schema

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_githubauth(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.schema.query.githubauth", y)
    yield y


##__________________________________________________________________||
def test_auth(app, mock_githubauth):

    query = '{ githubUser { login name avatarUrl } }'

    viewer = {
        "login": "octocat",
        "name": "monalisa octocat",
        "avatarUrl": "https://github.com/images/error/octocat_happy.gif"
    }
    mock_githubauth.get_user.return_value = viewer

    expected = {
        'githubUser': viewer
    }

    context = Context(headers=Headers({'Authorization': 'Bearer "token0123"'}))

    with app.app_context():
        schema = create_schema()
        client = Client(schema)
        result = client.execute(query, context_value=context)
        assert [mock.call('token0123')] == mock_githubauth.get_user.call_args_list
        assert {'data': expected} == result

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

from ..gql import DELETE_PRODUCT_TYPE

QEURY = '''
{
  allProductTypes {
    edges {
      node {
        name
        typeId
      }
    }
  }
}
'''

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [DELETE_PRODUCT_TYPE],
            {'variables': { 'typeId': 2 }},
        ],
        [[QEURY], {}],
        id='delete'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [DELETE_PRODUCT_TYPE],
            {'variables': { 'typeId': 12 }},
        ],
        [[QEURY], {}],
        id='error-nonexistent'
    ),
    pytest.param(
        [
            [DELETE_PRODUCT_TYPE],
            {'variables': { 'typeId': 1 }},
        ],
        [[QEURY], {}],
        id='error-unempty'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

##__________________________________________________________________||



import graphene
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import Product as ProductModel
from ..models import ProductRelation as ProductRelationModel
from ..models import ProductRelationType as ProductRelationTypeModel

from ..db.sa import sa
from ..db.backup import request_backup_db

from .connection import CountedConnection
from .filter_ import PFilterableConnectionField

##__________________________________________________________________||
class ProductRelation(SQLAlchemyObjectType):
    '''A relation from one product to another'''
    class Meta:
        model = ProductRelationModel
        interfaces = (graphene.relay.Node, )
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory

##__________________________________________________________________||
class CreateProductRelationInput(graphene.InputObjectType):
    '''An input to createProductRelation()'''
    type_id = graphene.Int(
        required=True,
        description=(
            'The typeId of the product relation type of the relation '
            'from "self" to the "other"'))
    self_product_id = graphene.Int(
        required=True,
        description=('The productId of the self product'))
    other_product_id = graphene.Int(
        required=True,
        description=('The productId of the other product'))

##__________________________________________________________________||
class CreateProductRelation(graphene.Mutation):
    '''Add relations between two products. The arguments only specify the relation
    from one product to the other. The reverse relation will be also added.

    '''
    class Arguments:
        input = CreateProductRelationInput(required=True)

    ok = graphene.Boolean()
    product_relation = graphene.Field(lambda: ProductRelation)

    def mutate(root, info, input):
        type_ = ProductRelationTypeModel.query.filter_by(type_id=input['type_id']).one()
        self_ = ProductModel.query.filter_by(product_id=input['self_product_id']).one()
        other = ProductModel.query.filter_by(product_id=input['other_product_id']).one()

        model = ProductRelationModel(type_=type_, self_=self_, other=other)

        sa.session.add(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return CreateProductRelation(product_relation=model, ok=ok)

class DeleteProductRelation(graphene.Mutation):
    '''Remove relations from two products.

    '''
    class Arguments:
        relation_id = graphene.Int(
            required=True,
            description=('The relationId of a relation. The reverse relation '
                         'will automatically be removed.'))

    ok = graphene.Boolean()

    def mutate(root, info, relation_id):
        model = ProductRelationModel.query.filter_by(relation_id=relation_id).one()
        sa.session.delete(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteProductRelation(ok=ok)

##__________________________________________________________________||



from flask import current_app
from sqlalchemy_utils import EncryptedType

from ..db.sa import sa

def encription_key():
    return current_app.config['SECRET_KEY']

##__________________________________________________________________||
class AdminAppToken(sa.Model):
    __tablename__ = 'admin_app_token'
    token_id = sa.Column(sa.Integer(), primary_key=True)
    token = sa.Column(EncryptedType(sa.Text(), key=encription_key))

##__________________________________________________________________||



"""Configure logging

"""
import logging
from logging.config import dictConfig
from pathlib import Path

##__________________________________________________________________||
_module_path = Path(__file__).resolve().parent.parent
# the path to the dir in which the module is installed,
# i.e., the one dir above the module path.
# e.g., /home/username/venv/lib/python3.8/site-packages

# https://docs.python.org/3/library/logging.html#logrecord-objects
_old_factory = logging.getLogRecordFactory()

def record_factory(*args, **kwargs):
    """replace the pathname (the full pathname of the source file) with
    the relative path of the source file in the package

    For example, replace
    "/home/username/venv/lib/python3.8/site-packages/acondbs/schema/auth.py"
    with
    "acondbs/schema/auth.py"

    """

    record = _old_factory(*args, **kwargs)
    try:
        record.pathname = Path(record.pathname).resolve().relative_to(_module_path)
    except:
        pass
    return record

logging.setLogRecordFactory(record_factory)

##__________________________________________________________________||
def configure_logging():
    """configure logging

    This function needs to be called early in the program, i.e., in
    the beginning of create_app() before app.logger is accessed. (It
    seems fine to be called after "app" is created.)

    https://flask.palletsprojects.com/en/1.1.x/logging/#basic-configuration

    """

    logger_name = __name__.split('.')[0] # i.e., "acondbs" (__name__ = "acondbs._logging")
    logger_level = 'DEBUG' # 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'
    dictConfig({
        'version': 1,
        'formatters': {'default': {
            'format': '{asctime} {levelname:>8s} {pathname}:{lineno}: {message}',
            'style': "{",
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }},
        'handlers': {'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }},
        'loggers': {
            logger_name: {
                'level': logger_level,
                'handlers': ['wsgi']
            }
        }
    })

##__________________________________________________________________||



import pytest

from sqlalchemy import exc

from acondbs.db.sa import sa
from acondbs.models import AdminAppToken

# __________________________________________________________________||
def test_token(app_empty):
    app = app_empty

    row = AdminAppToken(token="token123")

    with app.app_context():
        sa.session.add(row)
        sa.session.commit()

    with app.app_context():
        row = AdminAppToken.query.one()
        assert row.token == "token123"

# __________________________________________________________________||



import textwrap
from graphene.test import Client

import pytest
import unittest.mock as mock

from acondbs.schema.schema import create_schema

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_githubauth(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.schema.auth.githubauth", y)
    yield y


##__________________________________________________________________||
def test_auth(app, mock_githubauth):

    query = textwrap.dedent('''
        mutation GitHubAuth($code: String!) {
          githubAuth(code: $code) {
            authPayload {
              token
            }
          }
        }
    '''[1:])

    variables = { 'code': 'xyz' }

    mock_githubauth.get_token.return_value = 'token0123'

    expected = {
        'githubAuth': {
            'authPayload': {
                'token': 'token0123'
            }
        }
    }

    with app.app_context():
        schema = create_schema()
        client = Client(schema)
        result = client.execute(query, variables=variables, context_value={})
        assert {'data': expected} == result

##__________________________________________________________________||



import graphene
from graphql import GraphQLError

from ..models import AdminAppToken as AdminAppTokenModel
from ..misc import githubauth

from ..db.sa import sa

##__________________________________________________________________||
class OAuthAppInfo(graphene.ObjectType):
    client_id = graphene.String()
    authorize_url = graphene.String()
    token_url = graphene.String()
    redirect_uri = graphene.String()

##__________________________________________________________________||
class GitHubUser(graphene.ObjectType):
    login = graphene.String()
    name = graphene.String()
    avatarUrl = graphene.String() # Camel case so can easily be instantiated

##__________________________________________________________________||
class AuthPayload(graphene.ObjectType):
    token = graphene.String()

##__________________________________________________________________||
class GitHubAuth(graphene.Mutation):
    class Arguments:
        code = graphene.String(required=True)

    authPayload = graphene.Field(lambda: AuthPayload)

    def mutate(root, info, code):
        token = githubauth.get_token(code)
        if not token:
            raise GraphQLError('Unsuccessful to obtain the token')
        admin_token = AdminAppTokenModel.query.one()
        if not githubauth.is_member(user_token=token, admin_token=admin_token.token):
            raise GraphQLError('The user is not a member.')
        authPayload = AuthPayload(token=token)
        return GitHubAuth(authPayload=authPayload)

##__________________________________________________________________||
class StoreAdminAppToken(graphene.Mutation):
    class Arguments:
        code = graphene.String(required=True)

    ok = graphene.Boolean()

    def mutate(root, info, code):
        token = githubauth.get_token(code, admin=True)

        row = AdminAppTokenModel.query.one_or_none()
        if row:
            row.token = token
        else:
            row = AdminAppTokenModel(token=token)
            sa.session.add(row)
        sa.session.commit()
        ok = True
        return StoreAdminAppToken(ok=ok)


"""GitHub OAuth
"""

from flask import current_app
import requests

##__________________________________________________________________||
def get_token(code, admin=False):

    token_url = current_app.config['GITHUB_AUTH_TOKEN_URL']

    if admin:
        github_client_id = current_app.config['GITHUB_AUTH_ADMIN_CLIENT_ID']
        github_client_secret = current_app.config['GITHUB_AUTH_ADMIN_CLIENT_SECRET']
        redirect_uri = current_app.config['GITHUB_AUTH_ADMIN_REDIRECT_URI']
    else:
        github_client_id = current_app.config['GITHUB_AUTH_CLIENT_ID']
        github_client_secret = current_app.config['GITHUB_AUTH_CLIENT_SECRET']
        redirect_uri = current_app.config['GITHUB_AUTH_REDIRECT_URI']

    params = {
        'grant_type': 'authorization_code',
        'client_id': github_client_id,
        'client_secret': github_client_secret,
        'redirect_uri': redirect_uri,
        'code': code
    }

    headers = {
        'Accept': "application/vnd.github.v3+json, application/json",
    }

    r = requests.post(token_url, json=params, headers=headers)
    r = r.json()
    # type(r): <class 'dict'>
    # examples:
    #   success:
    #     r = {'access_token': 'xxx', 'token_type': 'bearer', 'scope': 'user'}
    #
    #   error:
    #     r = {
    #         'error': 'bad_verification_code',
    #         'error_description': 'The code passed is incorrect or expired.',
    #         'error_uri': 'https://docs.github.com/apps/managing-oauth-apps/troubleshooting-oauth-app-access-token-request-errors/#bad-verification-code'
    #     }

    token = r.get('access_token')
    return token

##__________________________________________________________________||
def get_user(token):
    headers = {
         'Authorization': 'token {}'.format(token)
    }

    json = {'query': '{ viewer { login name avatarUrl } }'}

    r = requests.post('https://api.github.com/graphql', json=json, headers=headers)

    r = r.json()
    # examples:
    #   success:
    #     r = {
    #         "data": {
    #             "viewer": {
    #                 "id": "MDQ6VXNlcjU4MzIzMQ==",
    #                 "name": "The Octocat",
    #                 "avatarUrl": "https://avatars3.githubusercontent.com/u/583231?u=a59fef2a493e2b67dd13754231daf220c82ba84d&v=4"
    #             }
    #         }
    #     }
    #
    #   error (bad credentials):
    #     r = {'message': 'Bad credentials', 'documentation_url': 'https://docs.github.com/graphql'}
    #
    #   error (query error):
    #     r = {'errors': [
    #         {
    #             'path': ['query', 'viewer', 'idii'],
    #             'extensions': {'code': 'undefinedField', 'typeName': 'User', 'fieldName': 'idii'},
    #             'locations': [{'line': 1, 'column': 12}],
    #             'message': "Field 'idii' doesn't exist on type 'User'"
    #         }
    #     ]}

    if 'errors' in r:
      raise Exception(r['errors'])

    user = r.get('data', {}).get('viewer')
    return user

##__________________________________________________________________||
def get_user_id(token):
    headers = {
         'Authorization': 'token {}'.format(token)
    }

    json = {'query': '{ viewer { id } }'}

    r = requests.post('https://api.github.com/graphql', json=json, headers=headers)

    r = r.json()
    # examples:
    #   success:
    #     r = {
    #         "data": {
    #             "viewer": {
    #                 "id": "MDQ6VXNlcjU4MzIzMQ=="
    #             }
    #         }
    #     }
    #
    #   error (bad credentials):
    #     r = {'message': 'Bad credentials', 'documentation_url': 'https://docs.github.com/graphql'}
    #
    #   error (query error):
    #     r = {'errors': [
    #         {
    #             'path': ['query', 'viewer', 'idii'],
    #             'extensions': {'code': 'undefinedField', 'typeName': 'User', 'fieldName': 'idii'},
    #             'locations': [{'line': 1, 'column': 12}],
    #             'message': "Field 'idii' doesn't exist on type 'User'"
    #         }
    #     ]}

    if 'errors' in r:
      raise Exception(r['errors'])

    ret = r.get('data', {}).get('viewer', {}).get('id')
    return ret

##__________________________________________________________________||
def get_org_member_ids(org_name, token):

    headers = {
         'Authorization': 'bearer {}'.format(token)
    }

    query = """
      query OrganizationMemberCount($org_login: String!) {
        organization(login: $org_login) {
          login
          membersWithRole {
            totalCount
          }
        }
      }
    """
    variables = { "org_login": org_name }
    json = {'query': query, 'variables': variables }
    r = requests.post('https://api.github.com/graphql', json=json, headers=headers)
    r = r.json()
    # examples:
    #   success:
    #     r = {
    #         'data': {
    #             'organization': {
    #                 'login': 'urban-octo-disco',
    #                 'membersWithRole': {'totalCount': 2}
    #             }
    #         }
    #     }
    #   error:
    #     r = {
    #         'data': {'organization': None},
    #         'errors': [{
    #             'extensions': {'saml_failure': False},
    #             'locations': [{'column': 11, 'line': 5}],
    #             'message': 'Although you appear to have the correct authorization '
    #                        'credentials, the `simonsobs` organization has enabled '
    #                        'OAuth App access restrictions, meaning that data '
    #                        'access to third-parties is limited. For more '
    #                        'information on these restrictions, including how to '
    #                        'whitelist this app, visit '
    #                        'https://docs.github.com/articles/restricting-access-to-your-organization-s-data/',
    #             'path': ['organization', 'membersWithRole'],
    #             'type': 'FORBIDDEN'
    #         }]
    #     }

    if 'errors' in r:
      raise Exception(r['errors'])

    nmembers = r['data']['organization']['membersWithRole']['totalCount']

    query = """
      query OrganizationMembers($org_login: String!, $first: Int!) {
        organization(login: $org_login) {
          login
          membersWithRole(first: $first) {
            edges {
              role
              node {
                id
                login
              }
            }
          }
        }
      }
    """
    variables = { "org_login": org_name, "first": nmembers }
    json = {'query': query, 'variables': variables }
    r = requests.post('https://api.github.com/graphql', json=json, headers=headers)
    r = r.json()

    # examples:
    #   success:
    #     r = {
    #         'data': {
    #             'organization': {
    #                 'login': 'urban-octo-disco',
    #                 'membersWithRole': {
    #                     'edges': [
    #                         {
    #                             'node': {
    #                                 'id': 'MDQ6VXNlcjEzODgwODE=',
    #                                 'login': 'TaiSakuma'
    #                             },
    #                             'role': 'MEMBER'
    #                         },
    #                         {
    #                             'node': {
    #                                 'id': 'MDQ6VXNlcjE1Njg1Njk3',
    #                                 'login': 'tai-sakuma'
    #                             },
    #                             'role': 'ADMIN'
    #                         }
    #                     ]}
    #             }}
    #     }
    #   error:
    #     r = {
    #         'data': { 'organization': None},
    #         'errors': [{
    #             'extensions': {'saml_failure': False},
    #             'locations': [{'column': 11, 'line': 5}],
    #             'message': 'Although you appear to have the correct authorization '
    #                        'credentials, the `simonsobs` organization has enabled '
    #                        'OAuth App access restrictions, meaning that data '
    #                        'access to third-parties is limited. For more '
    #                        'information on these restrictions, including how to '
    #                        'whitelist this app, visit '
    #                        'https://docs.github.com/articles/restricting-access-to-your-organization-s-data/',
    #             'path': ['organization', 'membersWithRole'],
    #             'type': 'FORBIDDEN'
    #         }]
    #     }

    from pprint import pprint
    pprint(r)
    if 'errors' in r:
      raise Exception(r['errors'])

    edges = r['data']['organization']['membersWithRole']['edges']
    print(edges)

    member_ids = [e['node']['id'] for e in edges]
    # e.g., ['MDQ6VXNlcjEzODgwODE=', 'MDQ6VXNlcjE1Njg1Njk3']
    # Base64 encoded. To decode, use base64.b64decode
    # [b'04:User1388081', b'04:User15685697']
    return member_ids

##__________________________________________________________________||
def is_member(user_token, admin_token):
    github_org = current_app.config['GITHUB_ORG']
    member_ids = get_org_member_ids(org_name=github_org, token=admin_token)
    user_id = get_user_id(user_token)
    return user_id in member_ids

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

from ..gql import UPDATE_PRODUCT

QEURY = '''
{
  allProducts {
    edges {
      node {
        name
      }
    }
  }
  allProductRelations {
    edges {
      node {
        type_ {
          name
        }
        self_ {
          name
        }
        other {
          name
        }
      }
    }
  }
  allProductFilePaths {
    edges {
      node {
        path
      }
    }
  }
}
'''


##__________________________________________________________________||
params = [
    pytest.param(
        [
            [UPDATE_PRODUCT],
            {'variables': {
                'productId': 1,
                'input': {
                    'contact': "new-contact",
                    'updatedBy': "updater",
                    'note': "- updated note 123"
                }
            }},
        ],
        [[QEURY], {}],
        id='update'
    ),
    pytest.param(
        [
            [UPDATE_PRODUCT],
            {'variables': {
                'productId': 1,
                'input': {
                    'updatedBy': "updater",
                    'paths': [
                        "site1:/path/to/map1",
                        "site2:/updated/way/map1",
                        "site4:/additional/map1"
                    ],
                }
            }},
        ],
        [[QEURY], {}],
        id='update-paths'
    ),
    pytest.param(
        [
            [UPDATE_PRODUCT],
            {'variables': {
                'productId': 1,
                'input': {
                    'updatedBy': "updater",
                    'paths': [ ],
                }
            }},
        ],
        [[QEURY], {}],
        id='delete-paths'
    ),
    pytest.param(
        [
            [UPDATE_PRODUCT],
            {'variables': {
                'productId': 5,
                'input': {
                    'updatedBy': "updater",
                    'relations' : [
                        {'typeId': 1, 'productId': 4 },
                        {'typeId': 1, 'productId': 2 }
                    ]
                }
            }},
        ],
        [[QEURY], {}],
        id='update-relations'
    ),
    pytest.param(
        [
            [UPDATE_PRODUCT],
            {'variables': {
                'productId': 5,
                'input': {
                    'updatedBy': "updater",
                    'relations' : [ ]
                }
            }},
        ],
        [[QEURY], {}],
        id='delete-relations'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)


##__________________________________________________________________||
params = [
    pytest.param(
        [
            [UPDATE_PRODUCT],
            {'variables': {
                'productId': 1,
                'input': {
                    'name': "new-name",
                }
            }},
        ],
        [[QEURY], {}],
        id='error-immutable-fields'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

##__________________________________________________________________||



import pytest

import datetime

from acondbs import create_app
from acondbs.db.ops import define_tables

from acondbs.db.sa import sa
from acondbs.models import (
    ProductType,
    Product,
    ProductRelationType,
    ProductRelation
    )

##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri ="sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y

@pytest.fixture
def app(app_empty):

    y = app_empty

    # Relation types:
    #   parent <-> child
    #   plaintiff <-> defendant

    # Relations:
    #   map1 -> beam1
    #     |        |
    #     +--------+---> beam2

    # create relation types
    Parent = ProductRelationType(
        type_id=1, name='parent', indef_article='a', singular='parent', plural='parents')
    Parent.reverse = ProductRelationType(
        type_id=2, name='child', indef_article='a',singular='child', plural='children')
    Plaintiff = ProductRelationType(
        type_id=3, name='plaintiff', indef_article='a', singular='plaintiff', plural='plaintiffs')
    Plaintiff.reverse = ProductRelationType(
        type_id=4, name='defendant', indef_article='a', singular='defendant', plural='defendants')


    # create product types
    Map = ProductType(type_id=1, name='map')
    Beam = ProductType(type_id=2, name='beam')

    # create products
    map1 = Product(product_id=1, name="map1", date_produced=datetime.date(2020, 2, 1), type_=Map)
    beam1 = Product(product_id=4, name="beam1", date_produced=datetime.date(2020, 2, 5), type_=Beam)
    beam2 = Product(product_id=5, name="beam2", date_produced=datetime.date(2020, 3, 4), type_=Beam)

    # create relations
    relation1 = ProductRelation(type_=Parent, self_=beam1, other=map1)
    relation2 = ProductRelation(type_=Parent, self_=beam2, other=map1)
    relation3 = ProductRelation(type_=Parent, self_=beam2, other=beam1)

    with y.app_context():
        sa.session.add(Parent)
        sa.session.add(Plaintiff)
        sa.session.commit()
    yield y

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

from ..gql import (
    FRAGMENT_PRODUCT_RELATION_CONNECTION,
    CREATE_PRODUCT_RELATION
    )

QEURY = '''
{
  allProductRelations {
    ...fragmentProductRelationConnection
  }
}
''' + FRAGMENT_PRODUCT_RELATION_CONNECTION

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION],
            { 'variables': {
                'input': {
                    'typeId': 1,
                    'selfProductId': 5,
                    'otherProductId': 1
                }
            }}
        ],
        [[QEURY], {}],
        id='create'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION],
            { 'variables': {
                'input': {
                    'typeId': 20,
                    'selfProductId': 5,
                    'otherProductId': 1
                }
            }}
        ],
        [[QEURY], {}],
        id='error-type_id-nonexistent'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION],
            { 'variables': {
                'input': {
                    'typeId': 1,
                    'selfProductId': 10,
                    'otherProductId': 1
                }
            }}
        ],
        [[QEURY], {}],
        id='error-self_product_id-nonexistent'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION],
            { 'variables': {
                'input': {
                    'typeId': 1,
                    'selfProductId': 5,
                    'otherProductId': 20
                }
            }}
        ],
        [[QEURY], {}],
        id='error-otheer_product_id-nonexistent'
    ),
    pytest.param(
        [
            [CREATE_PRODUCT_RELATION],
            { 'variables': {
                'input': {
                    'typeId': 2,
                    'selfProductId': 1,
                    'otherProductId': 4
                }
            }}
        ],
        [[QEURY], {}],
        id='error-duplicate'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

##__________________________________________________________________||




##__________________________________________________________________||
FRAGMENT_PRODUCT_RELATION_TYPE = '''
fragment fragmentProductRelationType on ProductRelationType {
  typeId
  name
  indefArticle
  singular
  plural
  reverse {
    typeId
    name
  }
  relations {
    edges {
      node {
        self_ {
          productId
          name
        }
        other {
          productId
          name
        }
      }
    }
  }
}
'''

FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION = '''
fragment fragmentProductRelationTypeConnection on ProductRelationTypeConnection {
  edges {
    node {
      ...fragmentProductRelationType
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE

##__________________________________________________________________||
CREATE_PRODUCT_RELATION_TYPES = '''
mutation CreateProductRelationTypes(
    $type: CreateProductRelationTypeInput!,
    $reverse: CreateProductRelationTypeInput,
    $selfReverse: Boolean) {
  createProductRelationTypes(type: $type, reverse: $reverse, selfReverse: $selfReverse) {
    ok
    productRelationType {
      ...fragmentProductRelationType
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE

DELETE_PRODUCT_RELATION_TYPES = '''
mutation DeleteProductRelationTypes($typeId: Int!) {
  deleteProductRelationTypes(typeId: $typeId) {
    ok
  }
}
'''

UPDATE_PRODUCT_RELATION_TYPE = '''
mutation UpdateProductRelationType(
    $typeId: Int!, $input: UpdateProductRelationTypeInput!) {
  updateProductRelationType(typeId: $typeId, input: $input) {
    ok
    productRelationType {
      ...fragmentProductRelationType
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE

##__________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

from ..gql import FRAGMENT_PRODUCT_RELATION

##__________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          {
            productRelation(relationId: 1) {
              ...fragmentProductRelation
            }
          }
        ''') + FRAGMENT_PRODUCT_RELATION,],
        {},
        id='type_id'
    ),
    pytest.param(
        [textwrap.dedent('''
          {
            productRelation(relationId: 222) {
              ...fragmentProductRelation
            }
          }
        ''') + FRAGMENT_PRODUCT_RELATION,],
        {},
        id='type_id-nonexistent'
    ),
]

##__________________________________________________________________||
@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

# __________________________________________________________________||



import pytest

import datetime

from acondbs import create_app
from acondbs.db.ops import define_tables

from acondbs.db.sa import sa
from acondbs.models import (
    ProductType,
    Product
    )

##__________________________________________________________________||
@pytest.fixture
def app_empty():
    database_uri ="sqlite:///:memory:"
    y = create_app(SQLALCHEMY_DATABASE_URI=database_uri)
    with y.app_context():
        define_tables()
    yield y

@pytest.fixture
def app(app_empty):

    y = app_empty

    # create product types
    Map = ProductType(type_id=1, name='map', order=2, indef_article='a',
                      singular='map', plural='maps', icon='mdi-map')
    Beam = ProductType(type_id=2, name='beam', order=1, indef_article='a',
                      singular='beam', plural='beams', icon='mdi-spotlight-beam')

    # create products
    map1 = Product(name="map1", date_produced=datetime.date(2020, 2, 1), type_=Map)
    map2 = Product(name="map2", date_produced=datetime.date(2020, 5, 10), type_=Map)
    map3 = Product(name="map3", date_produced=datetime.date(2020, 3, 3), type_=Map)

    with y.app_context():
        sa.session.add(Map)
        sa.session.add(Beam)
        sa.session.commit()
    yield y

##__________________________________________________________________||



from flask import Blueprint, request
from flask_graphql import GraphQLView

from .schema.schema import create_schema

##__________________________________________________________________||
# from flask import request
# class GraphQLView(GraphQLView):
#     def print_request(self):
#         import json
#         import textwrap
#         h = request.headers
#         h = str(h)
#         data_dict = json.loads(request.data)
#         m = '\n'.join([
#             textwrap.dedent('''
#             {}:
#             {}
#             ''').lstrip().format(k, textwrap.indent(str(v), '    ').rstrip()) for k, v in data_dict.items()]
#         )
#         msg = textwrap.dedent('''
#         - received header
#         {h}
#         - received query
#         {q}
#         --- end ---
#         ''').format(
#             h=textwrap.indent(h, '    '),
#             q=textwrap.indent(m, '    ')
#         )
#         print(msg)
#     def dispatch_request(self):
#         try:
#             self.print_request()
#         except BaseException as error:
#             import traceback
#             traceback.print_exc()
#         return super().dispatch_request()

##__________________________________________________________________||
bp = Blueprint('graphql', __name__)

schema_holder = []

class GraphQLViewW(GraphQLView):
    '''A wrapper.

    used to create schema in init_app() where app.config is accessible.


    The usual usage of GraphQLView in add_url_rule() as in the document
    (https://github.com/graphql-python/flask-graphql/tree/v2.0.1#usage) is as
    follows
      bp.add_url_rule('/graphql', view_func=GraphQLView.as_view('graphql', schema=schema, graphiql=True))
    The schema needs to be defined in the global scope.

    In this app, the schema is not available globally because its creation
    depends on the configuration. The configuration is not available globally.
    It is only available in the app context. Therefore, the schema is created
    in init_app(). Then it is place in the schema_holder.

    The __init__() will be called in each view from
    https://github.com/pallets/flask/blob/1.1.2/src/flask/views.py#L88

    In __init__(), schema is retrieved from schema_holder and is given to
    __init__() of GraphQLView.

    '''
    def __init__(self, **kwargs):
        kwargs['schema'] = schema_holder[0]
        super().__init__(**kwargs)


bp.add_url_rule('/graphql', view_func=GraphQLViewW.as_view('graphql', graphiql=True))

##__________________________________________________________________||
def init_app(app):
    with app.app_context():
        enable_mutation = not app.config.get('ACONDBS_SCHEME_MUTATION_DISABLE', False)
        schema = create_schema(enable_mutation=enable_mutation)
    schema_holder[:] = [schema]

    app.register_blueprint(bp)

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[app-admin] 1'] = {
    'data': {
        'oauthAppInfo': {
            'authorizeUrl': 'https://github.com/login/oauth/authorize',
            'clientId': 'client_id_admin_0123',
            'redirectUri': 'http://localhost:8080/admin/signin',
            'tokenUrl': 'https://github.com/login/oauth/access_token'
        }
    }
}

snapshots['test_schema[app] 1'] = {
    'data': {
        'oauthAppInfo': {
            'authorizeUrl': 'https://github.com/login/oauth/authorize',
            'clientId': 'client_id_0123456789',
            'redirectUri': 'http://localhost:8080/signin',
            'tokenUrl': 'https://github.com/login/oauth/access_token'
        }
    }
}



"""add a column to product_types

Revision ID: d8afc036053a
Revises: d7cac99bbffa
Create Date: 2020-05-22 13:02:07.024023

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd8afc036053a'
down_revision = 'd7cac99bbffa'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('product_types', sa.Column('order', sa.Integer(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('product_types', 'order')
    # ### end Alembic commands ###



from pathlib import Path
import pytest

from acondbs import create_app
from acondbs.db.ops import define_tables
from acondbs.db.ops import import_tables_from_csv_files
from acondbs.db.ops import export_db_to_dict_of_dict_list
from acondbs.db.ops import export_db_to_csv_files

from acondbs.models import AdminAppToken

from ...constants import SAMPLE_DIR

##__________________________________________________________________||
@pytest.fixture
def app():
    config_path = Path(SAMPLE_DIR, 'config.py')
    database_uri ="sqlite:///:memory:"
    app = create_app(config_path=config_path, SQLALCHEMY_DATABASE_URI=database_uri)
    with app.app_context():
        define_tables()
    yield app


##__________________________________________________________________||
def test_non_empty(app, snapshot):
    csvdir = Path(SAMPLE_DIR, 'csv')
    with app.app_context():
        import_tables_from_csv_files(csvdir)
    with app.app_context():
        snapshot.assert_match(export_db_to_dict_of_dict_list())

##__________________________________________________________________||
def testt_empty(app, tmpdir_factory, snapshot):
    csvdir = str(tmpdir_factory.mktemp('csv'))

    with app.app_context():
        export_db_to_csv_files(csvdir)

    with app.app_context():
        import_tables_from_csv_files(csvdir)

##__________________________________________________________________||
def test_encrypted_field(app):
    csvdir = Path(SAMPLE_DIR, 'csv')
    with app.app_context():
        import_tables_from_csv_files(csvdir)
        assert 'token123' == AdminAppToken.query.one().token

def test_how_to_encrypt_and_decrypt(app, snapshot):
    unencrypted = 'token123'
    encrypted = 'aKjGknYDHY39Z2xAaN7+sQ=='

    with app.app_context():
        secret_key = app.config['SECRET_KEY'] # 'secret_key_test_123'

    from sqlalchemy_utils.types.encrypted.encrypted_type import AesEngine
    engine = AesEngine()
    engine._update_key(secret_key)
    engine._set_padding_mechanism(None)

    assert engine.encrypt(unencrypted) == encrypted
    assert engine.decrypt(encrypted) == unencrypted

##__________________________________________________________________||



from pathlib import Path

import textwrap
from graphene.test import Client

import pytest
import unittest.mock as mock

from acondbs import create_app
from acondbs.db.ops import define_tables
from acondbs.schema.schema import create_schema
from acondbs.db.sa import sa
from acondbs.models import AdminAppToken

from ...constants import SAMPLE_DIR

##__________________________________________________________________||
STORE_ADMIN_APP_TOKEN = '''
mutation StoreAdminAppToken($code: String!) {
  storeAdminAppToken(code: $code) {
    ok
  }
}
'''

##__________________________________________________________________||
@pytest.fixture
def app():
    config_path = Path(SAMPLE_DIR, 'config.py')
    database_uri ="sqlite:///:memory:"
    app = create_app(config_path=config_path, SQLALCHEMY_DATABASE_URI=database_uri)
    with app.app_context():
        define_tables()
    yield app


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_githubauth(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.schema.auth.githubauth", y)
    yield y


##__________________________________________________________________||
def test_store_admin_app_token(app, mock_githubauth):

    query = STORE_ADMIN_APP_TOKEN
    variables = { 'code': 'xyz' }
    mock_githubauth.get_token.return_value = 'token0123'

    expected = {
        'storeAdminAppToken': {
            'ok': True
        }
    }

    with app.app_context():
        schema = create_schema()
        client = Client(schema)
        result = client.execute(query, variables=variables, context_value={})
        assert {'data': expected} == result

    with app.app_context():
        token = AdminAppToken.query.one()
        assert 'token0123' == token.token
        print(token.token)

def test_store_admin_app_token_update(app, mock_githubauth):

    with app.app_context():
        row = AdminAppToken(token='old_token_xyz')
        sa.session.add(row)
        sa.session.commit()

    query = STORE_ADMIN_APP_TOKEN
    variables = { 'code': 'xyz' }
    mock_githubauth.get_token.return_value = 'new_token_0123'

    expected = {
        'storeAdminAppToken': {
            'ok': True
        }
    }

    with app.app_context():
        schema = create_schema()
        client = Client(schema)
        result = client.execute(query, variables=variables, context_value={})
        assert {'data': expected} == result

    with app.app_context():
        token = AdminAppToken.query.one()
        assert 'new_token_0123' == token.token

##__________________________________________________________________||



import graphene
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..models import Product as ProductModel
from ..models import ProductRelation as ProductRelationModel
from ..models import ProductRelationType as ProductRelationTypeModel

from ..db.sa import sa
from ..db.backup import request_backup_db

from .connection import CountedConnection
from .filter_ import PFilterableConnectionField

##__________________________________________________________________||
class ProductRelationType(SQLAlchemyObjectType):
    '''A type of relations between products'''
    class Meta:
        model = ProductRelationTypeModel
        interfaces = (graphene.relay.Node, )
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory

##__________________________________________________________________||
class CommonInputFields:
    indef_article = graphene.String(
        description=('The indefinite article placed before the singular noun "'
                     'i.e., "a" or "an". '))
    singular = graphene.String(
        description=('The singular noun, the relation type name in singular.'))
    plural = graphene.String(
        description=('The plural noun, the relation type name in plural.'))

class CreateProductRelationTypeInput(graphene.InputObjectType, CommonInputFields):
    '''An input to createProductRelationTypes()'''
    name = graphene.String(
        required=True,
        description=('The name of the relation type'))

class UpdateProductRelationTypeInput(graphene.InputObjectType, CommonInputFields):
    '''An input to updateProductRelationType()'''

##__________________________________________________________________||
class CreateProductRelationTypes(graphene.Mutation):
    '''Create a pair of product relation types'''
    class Arguments:
        type = CreateProductRelationTypeInput(
            required=True,
            description=('A relation type'))
        reverse = CreateProductRelationTypeInput(
            description=('The reverse relation type'))
        self_reverse = graphene.Boolean(
            description=('true if the reverse type is the same'))

    ok = graphene.Boolean()
    product_relation_type = graphene.Field(lambda: ProductRelationType)

    def mutate(root, info, type, reverse=None, self_reverse=False):
        if self_reverse and reverse:
            from graphql import GraphQLError
            raise GraphQLError('"reverse" is given when "self_reverse" is True')
        model = ProductRelationTypeModel(**type)
        if self_reverse:
            model.reverse = model
        else:
            reverse_ = ProductRelationTypeModel(**reverse)
            model.reverse = reverse_
        sa.session.add(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return CreateProductRelationTypes(product_relation_type=model, ok=ok)

class UpdateProductRelationType(graphene.Mutation):
    '''Update a product relation type'''
    class Arguments:
        type_id = graphene.Int(
            required=True, description=('The typeId of the product relation type'))
        input = UpdateProductRelationTypeInput(required=True)

    ok = graphene.Boolean()
    product_relation_type = graphene.Field(lambda: ProductRelationType)

    def mutate(root, info, type_id, input):
        model = ProductRelationTypeModel.query.filter_by(type_id=type_id).one()
        for k, v in input.items():
            setattr(model, k, v)
        sa.session.commit()
        ok = True
        request_backup_db()
        return UpdateProductRelationType(product_relation_type=model, ok=ok)

class DeleteProductRelationTypes(graphene.Mutation):
    '''Delete a pair of product relation types'''
    class Arguments:
        type_id = graphene.Int(
            required=True,
            description=('The typeId of a relation type. The reverse relation '
                         'type will automatically be deleted.'))

    ok = graphene.Boolean()

    def mutate(root, info, type_id):
        model = ProductRelationTypeModel.query.filter_by(type_id=type_id).one()
        sa.session.delete(model)
        sa.session.commit()
        ok = True
        request_backup_db()
        return DeleteProductRelationTypes(ok=ok)

##__________________________________________________________________||



import pytest

from ..funcs import assert_mutation

from ..gql import CREATE_PRODUCT_TYPE

QEURY = '''
{
  allProductTypes {
    edges {
      node {
        name
      }
    }
  }
}
'''

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT_TYPE],
            {'variables': {
                'input': {
                  'name': "compass",
                  'order': 5,
                  'indefArticle': "a",
                  'singular': "compass",
                  'plural': "compasses",
                  'icon': "mdi-compass"
                }
            }},
        ],
        [[QEURY], {}],
        id='create'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_success(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=True)

##__________________________________________________________________||
params = [
    pytest.param(
        [
            [CREATE_PRODUCT_TYPE],
            {'variables': {
                'input': {
                  'name': "map",
                  'order': 5,
                  'indefArticle': "a",
                  'singular': "map",
                  'plural': "maps",
                  'icon': "mdi-map"
                }
            }},
        ],
        [[QEURY], {}],
        id='error-already-exist'
    ),
]

@pytest.mark.parametrize('mutation, query', params)
def test_schema_error(app, snapshot, mutation, query, mock_request_backup_db):
    assert_mutation(app, snapshot, mutation, query,
                    mock_request_backup_db, success=False)

##__________________________________________________________________||



import pytest
import textwrap

from ..funcs import assert_query

# __________________________________________________________________||
params = [
    pytest.param(
        [textwrap.dedent('''
          { webConfig {
              configId
              headTitle
              toolbarTitle
              devtoolLoadingstate
              productCreationDialog
              productUpdateDialog
              productDeletionDialog
            }
          }
          '''), ],
        {},
        id='query'
    ),
]

# __________________________________________________________________||


@pytest.mark.parametrize('args, kwags', params)
def test_schema(app, snapshot, args, kwags):
    assert_query(app, snapshot, [args, kwags])

# __________________________________________________________________||
